/* GNIO - GLib Network Layer of GIO
 *
 * Copyright (C) 2008 Christian Kellner, Samuel Cormier-Iijima
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Christian Kellner <gicmo@gnome.org>
 *          Samuel Cormier-Iijima <sciyoshi@gmail.com>
 */

#include <config.h>
#include <glib.h>
#include <gio/gio.h>

#include "gsocket.h"
#include "gnetworkinputstream.h"

G_DEFINE_TYPE (GNetworkInputStream, g_network_input_stream, G_TYPE_INPUT_STREAM);

enum
{
  PROP_0,
  PROP_SOCKET
};

struct _GNetworkInputStreamPrivate
{
  GSocket *socket;
};

static void
g_network_input_stream_get_property (GObject    *object,
                                     guint       prop_id,
                                     GValue     *value,
                                     GParamSpec *pspec)
{
  GNetworkInputStream *stream = G_NETWORK_INPUT_STREAM (object);

  switch (prop_id)
    {
      case PROP_SOCKET:
        g_value_set_object (value, stream->priv->socket);
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
g_network_input_stream_set_property (GObject      *object,
                                     guint         prop_id,
                                     const GValue *value,
                                     GParamSpec   *pspec)
{
  GNetworkInputStream *stream = G_NETWORK_INPUT_STREAM (object);

  switch (prop_id)
    {
      case PROP_SOCKET:
        stream->priv->socket = g_value_dup_object (value);
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
g_network_input_stream_finalize (GObject *object)
{
  if (G_OBJECT_CLASS (g_network_input_stream_parent_class)->finalize)
    (*G_OBJECT_CLASS (g_network_input_stream_parent_class)->finalize) (object);
}

static void
g_network_input_stream_dispose (GObject *object)
{
  if (G_OBJECT_CLASS (g_network_input_stream_parent_class)->dispose)
    (*G_OBJECT_CLASS (g_network_input_stream_parent_class)->dispose) (object);
}

static gssize
g_network_input_stream_read (GInputStream  *stream,
                             void          *buffer,
                             gsize          count,
                             GCancellable  *cancellable,
                             GError       **error)
{
  GNetworkInputStream *input_stream = G_NETWORK_INPUT_STREAM (stream);

  return g_socket_receive (input_stream->priv->socket, (gchar *) buffer, count, error);
}

static void
g_network_input_stream_class_init (GNetworkInputStreamClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GInputStreamClass *ginputstream_class = G_INPUT_STREAM_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GNetworkInputStreamPrivate));

  gobject_class->finalize = g_network_input_stream_finalize;
  gobject_class->dispose = g_network_input_stream_dispose;
  gobject_class->get_property = g_network_input_stream_get_property;
  gobject_class->set_property = g_network_input_stream_set_property;

  ginputstream_class->read_fn = g_network_input_stream_read;

  g_object_class_install_property (gobject_class, PROP_SOCKET,
                                   g_param_spec_object ("socket",
                                                        "socket",
                                                        "the socket that this stream wraps",
                                                        G_TYPE_SOCKET,
                                                        G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB | G_PARAM_STATIC_NICK));
}

static void
g_network_input_stream_init (GNetworkInputStream *stream)
{
  stream->priv = G_TYPE_INSTANCE_GET_PRIVATE (stream, G_TYPE_NETWORK_INPUT_STREAM, GNetworkInputStreamPrivate);

  stream->priv->socket = NULL;
}

GNetworkInputStream *
_g_network_input_stream_new (GSocket *socket)
{
  return G_NETWORK_INPUT_STREAM (g_object_new (G_TYPE_NETWORK_INPUT_STREAM, "socket", socket, NULL));
}
