/* GNIO - GLib Network Layer of GIO
 *
 * Copyright (C) 2008 Christian Kellner, Samuel Cormier-Iijima
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Christian Kellner <gicmo@gnome.org>
 *          Samuel Cormier-Iijima <sciyoshi@gmail.com>
 */

#include <config.h>
#include <glib.h>
#include <gio/gio.h>
#include <gnio/gnio.h>

#include <string.h>
#include <errno.h>

G_DEFINE_TYPE (GTcpClient, g_tcp_client, G_TYPE_OBJECT);

enum
{
  PROP_0,
  PROP_ADDRESS,
  PROP_HOSTNAME,
  PROP_PORT
};

struct _GTcpClientPrivate
{
  GInetSocketAddress *address;
  gchar              *hostname;
  gushort             port;
  GSocket            *socket;
};

static void
g_tcp_client_constructed (GObject *object)
{
  GTcpClient *client = G_TCP_CLIENT (object);

  if (client->priv->address)
    {
      // we've been constructed with an address, extract hostname+port
      client->priv->hostname = g_inet_address_to_string (g_inet_socket_address_get_address (client->priv->address));
      client->priv->port = g_inet_socket_address_get_port (client->priv->address);
      return;
    }
}

static void
g_tcp_client_get_property (GObject    *object,
                           guint       prop_id,
                           GValue     *value,
                           GParamSpec *pspec)
{
  GTcpClient *client = G_TCP_CLIENT (object);

  switch (prop_id)
    {
      case PROP_ADDRESS:
        g_value_set_object (value, client->priv->address);
        break;

      case PROP_HOSTNAME:
        g_value_set_string (value, client->priv->hostname);
        break;

      case PROP_PORT:
        g_value_set_uint (value, client->priv->port);
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
g_tcp_client_set_property (GObject      *object,
                           guint         prop_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
  GTcpClient *client = G_TCP_CLIENT (object);

  switch (prop_id)
    {
      case PROP_ADDRESS:
        // sink the address' floating reference
        client->priv->address = G_INET_SOCKET_ADDRESS (g_value_get_object (value));
        if (client->priv->address)
          g_object_ref_sink (client->priv->address);
        break;

      case PROP_HOSTNAME:
        client->priv->hostname = g_value_dup_string (value);
        break;

      case PROP_PORT:
        client->priv->port = g_value_get_uint (value);
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
g_tcp_client_finalize (GObject *object)
{
  GTcpClient *client = G_TCP_CLIENT (object);

  g_object_unref (client->priv->address);

  if (G_OBJECT_CLASS (g_tcp_client_parent_class)->finalize)
    (*G_OBJECT_CLASS (g_tcp_client_parent_class)->finalize) (object);
}

static void
g_tcp_client_dispose (GObject *object)
{
  GTcpClient *client = G_TCP_CLIENT (object);

  g_free (client->priv->hostname);

  if (G_OBJECT_CLASS (g_tcp_client_parent_class)->dispose)
    (*G_OBJECT_CLASS (g_tcp_client_parent_class)->dispose) (object);
}

static void
g_tcp_client_class_init (GTcpClientClass *klass)
{
  GObjectClass *gobject_class G_GNUC_UNUSED = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GTcpClientPrivate));

  gobject_class->finalize = g_tcp_client_finalize;
  gobject_class->dispose = g_tcp_client_dispose;
  gobject_class->constructed = g_tcp_client_constructed;
  gobject_class->set_property = g_tcp_client_set_property;
  gobject_class->get_property = g_tcp_client_get_property;

  g_object_class_install_property (gobject_class, PROP_ADDRESS,
                                   g_param_spec_object ("address",
                                                        "address",
                                                        "the remote address the socket will connect to",
                                                        G_TYPE_INET_SOCKET_ADDRESS,
                                                        G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB | G_PARAM_STATIC_NICK));

  g_object_class_install_property (gobject_class, PROP_HOSTNAME,
                                   g_param_spec_string ("hostname",
                                                        "hostname",
                                                        "the hostname of the remote address the socket will connect to",
                                                        NULL,
                                                        G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB | G_PARAM_STATIC_NICK));

  g_object_class_install_property (gobject_class, PROP_PORT,
                                   g_param_spec_uint ("port",
                                                      "port",
                                                      "the remote port the socket will connect to",
                                                      0,
                                                      G_MAXUSHORT,
                                                      0,
                                                      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB | G_PARAM_STATIC_NICK));
}

static void
g_tcp_client_init (GTcpClient *client)
{
  client->priv = G_TYPE_INSTANCE_GET_PRIVATE (client, G_TYPE_TCP_CLIENT, GTcpClientPrivate);

  client->priv->address = NULL;
  client->priv->hostname = NULL;
  client->priv->port = 0;
  client->priv->socket = NULL;
}

GTcpClient *
g_tcp_client_new (const gchar *hostname,
                  gushort      port)
{
  return G_TCP_CLIENT (g_object_new (G_TYPE_TCP_CLIENT, "hostname", hostname, "port", port, NULL));
}

GTcpClient *
g_tcp_client_new_from_address (GInetSocketAddress *address)
{
  return G_TCP_CLIENT (g_object_new (G_TYPE_TCP_CLIENT, "address", address, NULL));
}

GNetworkInputStream *
g_tcp_client_get_input_stream (GTcpClient *client)
{
  if (!client->priv->socket)
    return NULL;

  return _g_network_input_stream_new (client->priv->socket);
}

GNetworkOutputStream *
g_tcp_client_get_output_stream (GTcpClient *client)
{
  if (!client->priv->socket)
    return NULL;

  return _g_network_output_stream_new (client->priv->socket);
}

gboolean
g_tcp_client_connect (GTcpClient    *client,
                      GCancellable  *cancellable,
                      GError       **error)
{
  GInetAddress *address;

  g_return_val_if_fail (G_IS_TCP_CLIENT (client), FALSE);

  if (!client->priv->address)
    {
      // we've been constructed with just hostname+port, resolve
      GResolver *resolver = g_resolver_new ();

      address = g_resolver_resolve (resolver, client->priv->hostname, cancellable, error);

      if (!address)
        return FALSE;

      client->priv->address = g_inet_socket_address_new (address, client->priv->port);

      g_object_unref (resolver);

      g_object_ref_sink (client->priv->address);
    }
  else
    {
     address = g_inet_socket_address_get_address (client->priv->address);
    }

  if (G_IS_INET4_ADDRESS (address))
    client->priv->socket = g_socket_new (G_SOCKET_DOMAIN_INET, G_SOCKET_TYPE_STREAM, NULL, error);
  else if (G_IS_INET6_ADDRESS (address))
    client->priv->socket = g_socket_new (G_SOCKET_DOMAIN_INET6, G_SOCKET_TYPE_STREAM, NULL, error);
  else
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED, "unsupported address domain");
      return FALSE;
    }

  if (!client->priv->socket)
    return FALSE;

  if (g_cancellable_set_error_if_cancelled (cancellable, error))
    return FALSE;

  if (!g_socket_connect (client->priv->socket, G_SOCKET_ADDRESS (client->priv->address), error))
    return FALSE;

  return TRUE;
}

typedef struct {
  GAsyncReadyCallback  callback;
  GCancellable        *cancellable;
  gpointer             user_data;
  GTcpClient          *client;
  gchar                address_buffer[256];
  gsize                address_length;
} ConnectData;

/*
static gboolean
connect_callback (ConnectData *data,
                  GIOCondition condition,
                  gint fd)
{
  return FALSE;
}
*/

void
g_tcp_client_connect_async (GTcpClient          *client,
                            GCancellable        *cancellable,
                            GAsyncReadyCallback  callback,
                            gpointer             user_data)
{

}

gboolean
g_tcp_client_connect_finish (GTcpClient    *client,
                             GAsyncResult  *result,
                             GError       **error)
{
  GSimpleAsyncResult *simple;

  g_return_val_if_fail (G_IS_TCP_CLIENT (client), FALSE);

  simple = G_SIMPLE_ASYNC_RESULT (result);

  if (g_simple_async_result_propagate_error (simple, error))
    return FALSE;

  g_warn_if_fail (g_simple_async_result_get_source_tag (simple) == g_tcp_client_connect_async);

  return TRUE;
}

void
g_tcp_client_close (GTcpClient *tcp_client)
{
  g_return_if_fail (G_IS_TCP_CLIENT (tcp_client));
}
